/**
 * Licensed to the University Corporation for Advanced Internet
 * Development, Inc. (UCAID) under one or more contributor license
 * agreements. See the NOTICE file distributed with this work for
 * additional information regarding copyright ownership.
 *
 * UCAID licenses this file to you under the Apache License,
 * Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the
 * License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific
 * language governing permissions and limitations under the License.
 */

/**
 * @file xmltooling/ElementProxy.h
 * 
 * An XMLObject with an open content model 
 */

#ifndef __xmltooling_eleproxy_h__
#define __xmltooling_eleproxy_h__

#include <xmltooling/AttributeExtensibleXMLObject.h>
#include <xmltooling/ElementExtensibleXMLObject.h>

namespace xmltooling {

    /**
     * An XMLObject with an open content model.
     */
    class XMLTOOL_API ElementProxy : public virtual AttributeExtensibleXMLObject, public virtual ElementExtensibleXMLObject
    {
    protected:
        ElementProxy();

    public:
        virtual ~ElementProxy();
    };
    
};

#endif /* __xmltooling_eleproxy_h__ */
