/* xmltooling/config_pub.h.  Generated from config_pub.h.in by configure.  */
/* Define if C++ compiler supports covariant virtual methods. */
#define HAVE_COVARIANT_RETURNS 1

/* Define to 1 if C++ compiler supports nullptr keyword. */
/* #undef HAVE_NULLPTR */

/* Define to 1 if you have an STL implementation that supports useful string
   specialization. */
#define HAVE_GOOD_STL 1

/* Define to 1 if you have an STL implementation that supports
   std::iterator_traits. */
#define HAVE_ITERATOR_TRAITS 1

/* Define if log4shib library is used. */
#define XMLTOOLING_LOG4SHIB 1

/* Define if log4cpp library is used. */
/* #undef XMLTOOLING_LOG4CPP */

/* Define to 1 to disable XML-Security-dependent features. */
/* #undef XMLTOOLING_NO_XMLSEC */

/* Define to 1 if Xerces supports a compliant DOMLS API. */
#define XMLTOOLING_XERCESC_COMPLIANT_DOMLS 1

/* Define to 1 if Xerces DOM ID methods take extra parameter.  */
#define XMLTOOLING_XERCESC_BOOLSETIDATTRIBUTE 1

/* Define to 1 if Xerces has a 64-bit-safe API. */
#define XMLTOOLING_XERCESC_64BITSAFE 1

/* Define to 1 if you have the `xsecsize_t' type. */
/* #undef HAVE_XSECSIZE_T */
